------------------------------------------------------------------------------
-- Snowflake Map Script by NycholusV
------------------------------------------------------------------------------

include "MapEnums"
include "MapUtilities"
include "MountainsCliffs"
include "RiversLakes"
include "FeatureGenerator"
include "TerrainGenerator"
include "NaturalWonderGenerator"
include "ResourceGenerator"
include "AssignStartingPlots"

local g_iW, g_iH;
local g_iFlags;

------------------------------------------------------------------------------

function GenerateMap()

	-- Get globals
	g_iW, g_iH = Map.GetGridSize();
	g_iFlags = TerrainBuilder.GetFractalFlags();
	
	-- Generate the snowflake blueprint
	local args = {
		radius_center = 10,
		radius_wing = 17
	};
	continentTypes, plotTypes, protectionData, nwData = GenerateSnowflake(args);
	
	-- Configure natural wonder placement
	local args = {
		iW = g_iW,
		iH = g_iH,
		plotTypes = plotTypes,
		protectionData = protectionData,
		nwData = nwData
	};
	protectionData, nwData = ConfigureNaturalWonders(args);
	
	-- Erode the continents
	local args = {
		iW = g_iW,
		iH = g_iH,
		continentTypes = continentTypes,
		plotTypes = plotTypes,
		protectionData = protectionData,
		aContinentModifiers = {1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.333},
		water_plot = g_PLOT_TYPE_OCEAN,
		water_terrain = g_TERRAIN_TYPE_OCEAN,
		sea_level = MapConfiguration.GetValue("sea_level")
	};
	plotTypes = ErodeLandmass(args);
	
	-- Get lists of plots in each continent sorted by fractal-generated height
	local args = {
		iW = g_iW,
		iH = g_iH,
		iFlags = g_iFlags,
		continentTypes = continentTypes,
		plotTypes = plotTypes,
		grain_amount = 12
	};
	heightRankings = GetHeightRankings(args);
	
	-- Add hills and mountains based on height rankings
	local args = {
		continentTypes = continentTypes,
		plotTypes = plotTypes,
		protectionData = protectionData,
		heightRankings = heightRankings,
		world_age = MapConfiguration.GetValue("world_age")
	};
	plotTypes = AddHillsMountains(args);
	
	-- Generate terrain
	local args = {
		iW = g_iW,
		iH = g_iH,
		iFlags = g_iFlags,
		plotTypes = plotTypes,
		temperature = MapConfiguration.GetValue("temperature")
	};
	terrainTypes = GenerateTerrain(args);
	
	-- Apply terrain
	local args = {
		iW = g_iW,
		iH = g_iH,
		plotTypes = plotTypes,
		terrainTypes = terrainTypes
	};
	ApplyTerrain(args);
	
	-- Add natural wonders
	AreaBuilder.Recalculate();
	AddNaturalWonders(nwData);
	
	-- Add rivers (base game function)
	AreaBuilder.Recalculate();
	AddRivers();
	
	-- Use height rankings to add lakes to lowest tiles
	local args = {
		continentTypes = continentTypes,
		plotTypes = plotTypes,
		protectionData = protectionData,
		heightRankings = heightRankings,
		sea_level = MapConfiguration.GetValue("sea_level")
	};
	AddLakes(args);
	
	-- Formally assign continents to plots
	local args = {
		continentTypes = continentTypes
	};
	StampContinents(args);
	
	-- Add features (base game function)
	local args = {
		rainfall = MapConfiguration.GetValue("rainfall")
	};
	local featuregen = FeatureGenerator.Create(args);
	featuregen:AddFeatures();
	
	-- Dead sea quick-fix
	for index, data in pairs(protectionData) do
		local pPlot = Map.GetPlotByIndex(index);
		if data.canHaveFeature == false and pPlot:IsNaturalWonder() == false then
			TerrainBuilder.SetFeatureType(pPlot, -1);
		end
	end
	
	-- Add cliffs (base game function)
	AddCliffs(plotTypes, terrainTypes);
	
	-- Add resources (base game function)
	resourcesConfig = MapConfiguration.GetValue("resources");
	local args = {
		continentTypes = continentTypes,
		resources = resourcesConfig,
		luxuryDistribution = {
			{1,   2,  3,  4},
			{4,   5,  6,  7},
			{7,   8,  9, 10},
			{10, 11, 12, 13},
			{13, 14, 15, 16},
			{16, 17, 18,  1},
			{19, 20, 21, 22, 23, 24}
		}
	};
	local resGen = ResourceGenerator.Create(args);
	
	-- Get starting locations (base game function)
	local startConfig = MapConfiguration.GetValue("start");
	local args = {
		MIN_MAJOR_CIV_FERTILITY = 150,
		MIN_MINOR_CIV_FERTILITY = 50, 
		MIN_BARBARIAN_FERTILITY = 1,
		START_MIN_Y = 15,
		START_MAX_Y = 15,
		START_CONFIG = startConfig,
	};
	local start_plot_database = AssignStartingPlots.Create(args)
	
	-- Add tribal villages (base game function)
	local GoodyGen = AddGoodies(g_iW, g_iH);
	
end

------------------------------------------------------------------------------

function GenerateSnowflake(args)
	
	-- Unpack arguments
	local radius_center = args.radius_center;
	local radius_wing = args.radius_wing;
	
	-- Returns:
	local continentTypes = {};
	local plotTypes = {};
	local protectionData = {};
	local nwData = {};
	
	-- Start location ring must be multiple of two to ensure centering
	local start_ring = 2 * math.floor((radius_center + math.floor(radius_wing / 2)) / 2);
	
	-- Set all plots to defaults
	for i = 0, (g_iW * g_iH) - 1 do
		continentTypes[i] = 0;
		plotTypes[i] = g_PLOT_TYPE_OCEAN;
		protectionData[i] = {};
		protectionData[i].canHaveLake = true;
		protectionData[i].canHaveMountain = true;
		protectionData[i].canHaveMinorCiv = true;
		protectionData[i].canHaveOcean = true;
		protectionData[i].isMajorCivStart = false;
		protectionData[i].mustHaveOcean = false;
		local pPlot = Map.GetPlotByIndex(i);
		TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_OCEAN);
	end
	
	-- Get center plot
	local iX = math.floor(g_iW / 2);
	local iY = math.floor(g_iH / 2);
	local pPlot = Map.GetPlot(iX, iY);
	local index = pPlot:GetIndex();
	continentTypes[index] = 7;
	plotTypes[index] = g_PLOT_TYPE_LAND;
	TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_SNOW);
	local row = {
		index = index,
		coastDir = -1
	};
	table.insert(nwData, row);
	
	-- Assign plots to continents
	for ring = 1, radius_center + radius_wing do
		local center = (ring + 2) / 2;
		pPlot = Map.GetAdjacentPlot(iX, iY, 4);
		iX = pPlot:GetX();
		iY = pPlot:GetY();
		index = pPlot:GetIndex();
		-- Navigate around the ring
		for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
			for i = 1, ring do
				pPlot = Map.GetAdjacentPlot(iX, iY, direction);
				iX = pPlot:GetX();
				iY = pPlot:GetY();
				index = pPlot:GetIndex();
				-- Is this plot part of the center continent?
				if ring <= radius_center then
					continentTypes[index] = 7;
					plotTypes[index] = g_PLOT_TYPE_LAND;
					TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_SNOW);
					-- Determine natural wonder placement
					if (ring == 6) and (i == 6) and (direction % 2 == 0) then
						local row = {
							index = index,
							coastDir = (direction + 5) % 6
						};
						table.insert(nwData, row);
					end
				-- Add plot to respective wing continent excluding boundaries
				elseif i < ring then
					continentTypes[index] = direction + 1;
					plotTypes[index] = g_PLOT_TYPE_LAND;
					TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_SNOW);
				end
				-- Determine protection
				if ring <= start_ring then
					if i == center and ring == start_ring then
						protectionData[index].isMajorCivStart = true;
					end
					if i > (center - 3) and i < (center + 3) then
						protectionData[index].canHaveMinorCiv = false;
						if i > (center - 2) and i < (center + 2) then
							protectionData[index].canHaveOcean = false;
							if i > (center - 1) and i < (center + 1) then
								protectionData[index].canHaveLake = false;
								protectionData[index].canHaveMountain = false;
							end
						end
					end
				end
			end
		end
	end
	
	return continentTypes, plotTypes, protectionData, nwData;
	
end

------------------------------------------------------------------------------

-- "NV_MapUtil_NaturalWonders.lua" (to be separated)
-- VFS system currently does not allow inclusion of imported scripts

------------------------------------------------------------------------------

-- Missing enums
g_FEATURE_EYJAFJALLAJOKULL = GetGameInfoIndex("Features", "FEATURE_EYJAFJALLAJOKULL");
g_FEATURE_LYSEFJORDEN = GetGameInfoIndex("Features", "FEATURE_LYSEFJORDEN");
g_FEATURE_GIANTS_CAUSEWAY = GetGameInfoIndex("Features", "FEATURE_GIANTS_CAUSEWAY");

-- special: Unusual NW placements that need to be separately accommodated for
-- nw_size: Number of tiles the NW consists of
-- east_west: Two-tile wonders that require an east-west orientation
-- anti_mountain = 0: Protect the NW plots from becoming mountains
-- anti_mountain = 1: Protect the NW and adjacent plots from becoming mountains
-- NOTE: All non-special NWs are non-coastal and therefore adjacent tiles are erosion-protected
g_nwProtections = {
	[g_FEATURE_BARRIER_REEF] 		=	{special = "BARRIER_REEF"},					--  6
	[g_FEATURE_CLIFFS_DOVER] 		=	{special = "CLIFFS_DOVER"},					--  7
	[g_FEATURE_CRATER_LAKE]			=	{nw_size = 1,		anti_mountain = 0},		--  8
	[g_FEATURE_DEAD_SEA]			=	{nw_size = 2,		anti_mountain = 1},		--  9
	[g_FEATURE_EVEREST]				=	{nw_size = 3,		anti_mountain = 1},		-- 10
	[g_FEATURE_GALAPAGOS]			=	{special = "GALAPAGOS"},					-- 11
	[g_FEATURE_KILIMANJARO]			=	{nw_size = 1,		anti_mountain = 1},		-- 12
	[g_FEATURE_PANTANAL]			=	{nw_size = 4,		anti_mountain = 0},		-- 13
	[g_FEATURE_PIOPIOTAHI]			=	{special = "PIOPIOTAHI"},					-- 14
	[g_FEATURE_TORRES_DEL_PAINE]	=	{east_west = true,	anti_mountain = 1},		-- 15
	[g_FEATURE_TSINGY]				=	{nw_size = 1,		anti_mountain = 1},		-- 16
	[g_FEATURE_YOSEMITE]			=	{east_west = true,	anti_mountain = 1},		-- 17
	[g_FEATURE_EYJAFJALLAJOKULL]	=	{east_west = true,	anti_mountain = 1},		-- 18
	[g_FEATURE_LYSEFJORDEN]			=	{special = "PIOPIOTAHI"},					-- 19
	[g_FEATURE_GIANTS_CAUSEWAY]		=	{special = "GIANTS_CAUSEWAY"}				-- 20
};

-- Terrain priorities
g_aaTerrainPriorities = {
	[g_TERRAIN_TYPE_GRASS] = {
		g_TERRAIN_TYPE_GRASS,
		g_TERRAIN_TYPE_PLAINS,
		g_TERRAIN_TYPE_TUNDRA,
		g_TERRAIN_TYPE_DESERT,
		g_TERRAIN_TYPE_SNOW
	},
	[g_TERRAIN_TYPE_PLAINS] = {
		g_TERRAIN_TYPE_PLAINS,
		g_TERRAIN_TYPE_GRASS,
		g_TERRAIN_TYPE_DESERT,
		g_TERRAIN_TYPE_TUNDRA,
		g_TERRAIN_TYPE_SNOW
	},
	[g_TERRAIN_TYPE_DESERT] = {
		g_TERRAIN_TYPE_DESERT,
		g_TERRAIN_TYPE_PLAINS,
		g_TERRAIN_TYPE_GRASS,
		g_TERRAIN_TYPE_TUNDRA,
		g_TERRAIN_TYPE_SNOW
	},
	[g_TERRAIN_TYPE_TUNDRA] = {
		g_TERRAIN_TYPE_TUNDRA,
		g_TERRAIN_TYPE_SNOW,
		g_TERRAIN_TYPE_PLAINS,
		g_TERRAIN_TYPE_GRASS,
		g_TERRAIN_TYPE_DESERT
	},
	[g_TERRAIN_TYPE_SNOW] = {
		g_TERRAIN_TYPE_SNOW,
		g_TERRAIN_TYPE_TUNDRA,
		g_TERRAIN_TYPE_PLAINS,
		g_TERRAIN_TYPE_GRASS,
		g_TERRAIN_TYPE_DESERT
	}
};

-- Plot priorities
g_aaPlotPriorities = {
	[g_PLOT_TYPE_LAND] = {
		g_PLOT_TYPE_LAND,
		g_PLOT_TYPE_HILLS,
		g_PLOT_TYPE_MOUNTAIN
	},
	[g_PLOT_TYPE_HILLS] = {
		g_PLOT_TYPE_HILLS,
		g_PLOT_TYPE_LAND,
		g_PLOT_TYPE_MOUNTAIN
	},
	[g_PLOT_TYPE_MOUNTAIN] = {
		g_PLOT_TYPE_MOUNTAIN,
		g_PLOT_TYPE_HILLS,
		g_PLOT_TYPE_LAND
	}
};

-- Flattened terrain types
g_aTerrainTypesFlat = {
	[g_TERRAIN_TYPE_GRASS] = g_TERRAIN_TYPE_GRASS,
	[g_TERRAIN_TYPE_GRASS_HILLS] = g_TERRAIN_TYPE_GRASS,
	[g_TERRAIN_TYPE_GRASS_MOUNTAIN] = g_TERRAIN_TYPE_GRASS,
	[g_TERRAIN_TYPE_PLAINS] = g_TERRAIN_TYPE_PLAINS,
	[g_TERRAIN_TYPE_PLAINS_HILLS] = g_TERRAIN_TYPE_PLAINS,
	[g_TERRAIN_TYPE_PLAINS_MOUNTAIN] = g_TERRAIN_TYPE_PLAINS,
	[g_TERRAIN_TYPE_DESERT] = g_TERRAIN_TYPE_DESERT,
	[g_TERRAIN_TYPE_DESERT_HILLS] = g_TERRAIN_TYPE_DESERT,
	[g_TERRAIN_TYPE_DESERT_MOUNTAIN] = g_TERRAIN_TYPE_DESERT,
	[g_TERRAIN_TYPE_TUNDRA] = g_TERRAIN_TYPE_TUNDRA,
	[g_TERRAIN_TYPE_TUNDRA_HILLS] = g_TERRAIN_TYPE_TUNDRA,
	[g_TERRAIN_TYPE_TUNDRA_MOUNTAIN] = g_TERRAIN_TYPE_TUNDRA,
	[g_TERRAIN_TYPE_SNOW] = g_TERRAIN_TYPE_SNOW,
	[g_TERRAIN_TYPE_SNOW_HILLS] = g_TERRAIN_TYPE_SNOW,
	[g_TERRAIN_TYPE_SNOW_MOUNTAIN] = g_TERRAIN_TYPE_SNOW
};

------------------------------------------------------------------------------

function AddNaturalWonders(nwData)
	
	print("Adding natural wonders");
	for i, data in ipairs(nwData) do
		-- Set the plot to an appropriate terrain type
		for iPlot, index in ipairs(data.plots) do
			print("Getting plot at index ", index);
			local pPlot = Map.GetPlotByIndex(index);
			local terrainTypeNew = AssimilateTerrain(index, data.validTerrains);
			TerrainBuilder.SetTerrainType(pPlot, terrainTypeNew);
			-- Set all adjacent plots to an appropriate terrain type
			for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
				local pAdjPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), direction);
				if pAdjPlot:IsNaturalWonder() == false then
					local adjIndex = pAdjPlot:GetIndex();
					print("Assimilating terrain for adjacent plot at index ", adjIndex);
					local adjTerrainTypeNew = AssimilateTerrain(adjIndex, data.adjacentTerrains);
					TerrainBuilder.SetTerrainType(pAdjPlot, adjTerrainTypeNew);
				end
			end
		end
		local index = data.index;
		local pPlot = Map.GetPlotByIndex(index);
		print("Placing natural wonder tile of feature type ", data.featureType);
		-- Valid feature test
		local customPlacement = GameInfo.Features[data.featureType].CustomPlacement;
		if customPlacement == nil then
			print("Can have wonder test: ", TerrainBuilder.CanHaveFeature(pPlot, data.featureType, false));
			TerrainBuilder.SetMultiPlotFeatureType(data.plots, data.featureType);
		else
			print("Can have wonder test: ", CustomCanHaveFeature(pPlot, data.featureType));
			TerrainBuilder.SetMultiPlotFeatureType(data.plots, data.featureType);
			for iIndex, nwPlotIndex in ipairs(data.plots) do
				SetNaturalCliff(nwPlotIndex);
			end
		end
		print("Natural wonder placed, attempting terrain reset");
		ResetTerrain(index);
		print("Attempting terrain reset for adjacent plots");
		for dx = -2, 2 do
			for dy = -2,2 do
				local otherPlot = Map.GetPlotXY(pPlot:GetX(), pPlot:GetY(), dx, dy, 2);
				if(otherPlot) then
					if(otherPlot:IsNaturalWonder() == true) then
						ResetTerrain(otherPlot:GetIndex());
					end
				end
			end
		end
		print("Success");
	end
	print("Successfully added natural wonders");
	
end

------------------------------------------------------------------------------

function ApplyNaturalWonderProtection(plotTypes, protectionData, data)
	
	local index = data.index;
	local pPlot = Map.GetPlotByIndex(index);
	local iX = pPlot:GetX();
	local iY = pPlot:GetY();
	local aProtections = g_nwProtections[data.featureType];
	
	print("Attempting to place natural wonder");
	print("NW plot is ", iX, iY);
	
	if aProtections.special == nil then
		print("Wonder placement is standard");
		-- Select plots that the NW tiles will be placed on
		data.plots = {};
		table.insert(data.plots, index);
		-- Does the NW require an east-west placement?
		if aProtections.east_west == true then
			local pAdjPlot = Map.GetAdjacentPlot(iX, iY, 1);
			local adjIndex = pAdjPlot:GetIndex();
			table.insert(data.plots, adjIndex);
			-- Yosemite quick-fix
			if data.featureType == g_FEATURE_YOSEMITE then
				pAdjPlot = Map.GetAdjacentPlot(iX, iY, 5);
				TerrainBuilder.SetFeatureType(pAdjPlot, g_FEATURE_FOREST);
			end
		elseif g_nwProtections[data.featureType].nw_size > 1 then
			local rand = TerrainBuilder.GetRandomNumber(6, "Random Direction");
			for rotation = 0, g_nwProtections[data.featureType].nw_size - 2 do
				local pAdjPlot = Map.GetAdjacentPlot(iX, iY, (rand + rotation) % 6);
				local adjIndex = pAdjPlot:GetIndex();
				table.insert(data.plots, adjIndex);
			end
		end
		-- Protect plots
		for i, index in ipairs(data.plots) do
			pPlot = Map.GetPlotByIndex(index);
			protectionData[index].canHaveMountain = false;
			protectionData[index].canHaveOcean = false;
			protectionData[index].canHaveLake = false;
			for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
				local pAdjPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), direction);
				local adjIndex = pAdjPlot:GetIndex();
				protectionData[adjIndex].canHaveOcean = false;
				protectionData[adjIndex].canHaveLake = false;
				if aProtections.anti_mountain == 1 then
					protectionData[adjIndex].canHaveMountain = false;
				end
				-- Dead sea quick-fix
				if data.featureType == g_FEATURE_DEAD_SEA then
					protectionData[index].canHaveFeature = false;
				end
			end
		end
		return;
	end
	
	if aProtections.special == "BARRIER_REEF" then
	print("Wonder placement is BARRIER_REEF");
		-- Get plots to place tiles on
		data.plots = {};
		table.insert(data.plots, index);
		protectionData[index].mustHaveOcean = true;
		local rand = TerrainBuilder.GetRandomNumber(6, "Random Direction");
		local pAdjPlot = Map.GetAdjacentPlot(iX, iY, rand);
		local adjIndex = pAdjPlot:GetIndex();
		table.insert(data.plots, adjIndex);
		protectionData[adjIndex].mustHaveOcean = true;
		-- Mandate ocean
		while plotTypes[index] ~= g_PLOT_TYPE_OCEAN do
			protectionData[index].mustHaveOcean = true;
			pPlot = Map.GetAdjacentPlot(iX, iY, data.coastDir);
			index = pPlot:GetIndex();
			iX = pPlot:GetX();
			iY = pPlot:GetY();
		end
		-- Protect at least one adjacent land tile from erosion for each tile
		for i, index in ipairs(data.plots) do
			local aValidLandPlots = {};
			local nValidLandPlots = 0;
			pPlot = Map.GetPlotByIndex(index);
			iX = pPlot:GetX();
			iY = pPlot:GetY();
			for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
				pAdjPlot = Map.GetAdjacentPlot(iX, iY, direction);
				adjIndex = pAdjPlot:GetIndex();
				if protectionData[adjIndex].mustHaveOcean == false then
					table.insert(aValidLandPlots, adjIndex);
					nValidLandPlots = nValidLandPlots + 1;
				end
			end
			local rand = 1 + TerrainBuilder.GetRandomNumber(nValidLandPlots, "Random Plot");
			protectionData[aValidLandPlots[rand]].canHaveOcean = false;
			protectionData[aValidLandPlots[rand]].canHaveLake = false;
		end
		return;
	end
	
	if aProtections.special == "CLIFFS_DOVER" then
	print("Wonder placement is CLIFFS_DOVER");
		-- Get plots to place tiles on
		data.plots = {};
		table.insert(data.plots, index);
		protectionData[index].canHaveOcean = false;
		protectionData[index].canHaveLake = false;
		protectionData[index].canHaveMountain = false;
		-- Starting plot is always westernmost
		local aDirections = {};
		local nDirections = 0;
		for direction = 0, 2 do
			if direction ~= data.coastDir then
				table.insert(aDirections, direction);
				nDirections = nDirections + 1;
			end
		end
		local rand = 1 + TerrainBuilder.GetRandomNumber(nDirections, "Random Direction");
		local direction = aDirections[rand];
		local pAdjPlot = Map.GetAdjacentPlot(iX, iY, direction);
		local adjIndex = pAdjPlot:GetIndex();
		table.insert(data.plots, adjIndex);
		protectionData[adjIndex].canHaveOcean = false;
		protectionData[adjIndex].canHaveLake = false;
		protectionData[adjIndex].canHaveMountain = false;
		-- Mandate ocean
		pAdjPlot = Map.GetAdjacentPlot(iX, iY, data.coastDir);
		adjIndex = pAdjPlot:GetIndex();
		while plotTypes[adjIndex] ~= g_PLOT_TYPE_OCEAN do
			protectionData[adjIndex].mustHaveOcean = true;
			pAdjPlot = Map.GetAdjacentPlot(pAdjPlot:GetX(), pAdjPlot:GetY(), data.coastDir);
			adjIndex = pAdjPlot:GetIndex();
		end
		-- Coastal edges appear on the westernmost (original) plot
		direction = (direction + 1) % 6;
		pAdjPlot = Map.GetAdjacentPlot(iX, iY, direction);
		adjIndex = pAdjPlot:GetIndex();
		local coastalEdges = 0;
		while protectionData[adjIndex].mustHaveOcean == false or coastalEdges < 2 do
			protectionData[adjIndex].mustHaveOcean = true;
			direction = (direction + 1) % 6;
			pAdjPlot = Map.GetAdjacentPlot(iX, iY, direction);
			adjIndex = pAdjPlot:GetIndex();
			coastalEdges = coastalEdges + 1;
		end
		return;
	end
	
	if aProtections.special == "GALAPAGOS" then
		print("Wonder placement is GALAPAGOS");
		-- Get plots to place tiles on
		data.plots = {};
		table.insert(data.plots, index);
		protectionData[index].mustHaveOcean = true;
		local rand = TerrainBuilder.GetRandomNumber(6, "Random Direction");
		local pAdjPlot = Map.GetAdjacentPlot(iX, iY, rand);
		local adjIndex = pAdjPlot:GetIndex();
		table.insert(data.plots, adjIndex);
		protectionData[index].mustHaveOcean = true;
		-- Mandate ocean
		pAdjPlot = Map.GetAdjacentPlot(iX, iY, data.coastDir);
		adjIndex = pAdjPlot:GetIndex();
		while plotTypes[adjIndex] ~= g_PLOT_TYPE_OCEAN do
			protectionData[adjIndex].mustHaveOcean = true;
			pAdjPlot = Map.GetAdjacentPlot(pAdjPlot:GetX(), pAdjPlot:GetY(), data.coastDir);
			adjIndex = pAdjPlot:GetIndex();
		end
		-- Mandate all adjacent plots to be ocean
		for i, index in ipairs(data.plots) do
			pPlot = Map.GetPlotByIndex(index);
			for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
				pAdjPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), direction);
				adjIndex = pAdjPlot:GetIndex();
				protectionData[adjIndex].mustHaveOcean = true;
			end
		end
		-- Protect one land tile two plots away
		index = data.plots[1];
		pPlot = Map.GetPlotByIndex(index);
		local directions = {};
		for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
			if direction ~= rand then
				table.insert(directions, direction);
			end
		end
		rand = TerrainBuilder.GetRandomNumber(DirectionTypes.NUM_DIRECTION_TYPES - 2, "Random Direction");
		local direction = directions[rand];
		pAdjPlot = Map.GetAdjacentPlot(iX, iY, direction);
		adjIndex = pAdjPlot:GetIndex();
		for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
			local pAdjNextPlot = Map.GetAdjacentPlot(pAdjPlot:GetX(), pAdjPlot:GetY(), direction);
			local adjNextIndex = pAdjNextPlot:GetIndex();
			if protectionData[adjNextIndex].mustHaveOcean ~= true then
				protectionData[adjNextIndex].canHaveOcean = false;
				protectionData[adjNextIndex].canHaveLake = false;
				break;
			end
		end
		return;
	end
	
	if aProtections.special == "PIOPIOTAHI" then
		print("Wonder placement is PIOPIOTAHI");
		-- Starting plot is inland
		data.plots = {};
		table.insert(data.plots, index);
		protectionData[index].canHaveOcean = false;
		protectionData[index].canHaveLake = false;
		protectionData[index].canHaveMountain = false;
		for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
			local pAdjPlot = Map.GetAdjacentPlot(iX, iY, direction);
			local adjIndex = pAdjPlot:GetIndex();
			protectionData[adjIndex].canHaveOcean = false;
			protectionData[adjIndex].canHaveLake = false;
		end
		-- Get second plot (in coastal direction)
		local pAdjPlot = Map.GetAdjacentPlot(iX, iY, data.coastDir);
		local adjIndex = pAdjPlot:GetIndex();
		table.insert(data.plots, adjIndex);
		protectionData[adjIndex].canHaveOcean = false;
		protectionData[adjIndex].canHaveLake = false;
		protectionData[adjIndex].canHaveMountain = false;
		-- Choose third plot
		local rand = 1 + TerrainBuilder.GetRandomNumber(2, "Random Direction");
		local pAdjNextPlot = Map.GetAdjacentPlot(pAdjPlot:GetX(), pAdjPlot:GetY(), (data.coastDir + 2 * rand) % 6);
		local adjNextIndex = pAdjNextPlot:GetIndex();
		table.insert(data.plots, adjNextIndex);
		protectionData[adjNextIndex].canHaveOcean = false;
		protectionData[adjNextIndex].canHaveLake = false;
		protectionData[adjNextIndex].canHaveMountain = false;
		-- Mandate coastal edges
		pAdjPlot = Map.GetAdjacentPlot(pAdjPlot:GetX(), pAdjPlot:GetY(), data.coastDir);
		adjIndex = pAdjPlot:GetIndex();
		protectionData[adjIndex].mustHaveOcean = true;
		pAdjNextPlot = Map.GetAdjacentPlot(pAdjPlot:GetX(), pAdjPlot:GetY(), (data.coastDir + 2 * rand) % 6);
		adjNextIndex = pAdjNextPlot:GetIndex();
		protectionData[adjNextIndex].mustHaveOcean = true;
		pAdjNextPlot = Map.GetAdjacentPlot(pAdjNextPlot:GetX(), pAdjNextPlot:GetY(), (data.coastDir + 2 * rand) % 6);
		adjNextIndex = pAdjNextPlot:GetIndex();
		protectionData[adjNextIndex].mustHaveOcean = true;
		-- Mandate ocean
		pAdjNextPlot = Map.GetAdjacentPlot(pAdjPlot:GetX(), pAdjPlot:GetY(), data.coastDir);
		adjNextPlotIndex = pAdjNextPlot:GetIndex();
		while plotTypes[adjNextPlotIndex] ~= g_PLOT_TYPE_OCEAN do
			protectionData[adjNextPlotIndex].mustHaveOcean = true;
			pAdjNextPlot = Map.GetAdjacentPlot(pAdjNextPlot:GetX(), pAdjNextPlot:GetY(), data.coastDir);
			adjNextPlotIndex = pAdjNextPlot:GetIndex();
		end
		return;
	end
	
	if aProtections.special == "GIANTS_CAUSEWAY" then
		print("Wonder placement is GIANTS_CAUSEWAY");
		-- Starting plot is land
		data.plots = {};
		table.insert(data.plots, index);
		protectionData[index].canHaveOcean = false;
		protectionData[index].canHaveLake = false;
		protectionData[index].canHaveMountain = false;
		-- Mandate ocean
		local pAdjPlot = Map.GetAdjacentPlot(iX, iY, data.coastDir);
		local adjIndex = pAdjPlot:GetIndex();
		while plotTypes[adjIndex] ~= g_PLOT_TYPE_OCEAN do
			protectionData[adjIndex].mustHaveOcean = true;
			pAdjPlot = Map.GetAdjacentPlot(pAdjPlot:GetX(), pAdjPlot:GetY(), data.coastDir);
			adjIndex = pAdjPlot:GetIndex();
		end
		-- Select plot for coast tile (not plot further from coast)
		local direction = (((data.coastDir + 3) % 6) + 1 + TerrainBuilder.GetRandomNumber(5, "Random Plot")) % 6;
		pAdjPlot = Map.GetAdjacentPlot(iX, iY, direction);
		adjIndex = pAdjPlot:GetIndex();
		table.insert(data.plots, adjIndex);
		protectionData[adjIndex].mustHaveOcean = true;
		-- Navigate around to ocean
		local rotation;
		if data.coastDir == direction then
			table.insert(data.plots, adjIndex);
			protectionData[adjIndex].mustHaveOcean = true;
			return;
		elseif math.abs((data.coastDir - direction) % 2) > 3 then
			rotation = 5;
		elseif math.abs((data.coastDir - direction) % 2) < 3 then
			rotation = 1;
		else
			local rand = TerrainBuilder.GetRandomNumber(2, "Random Rotation");
			rotation = 1 + 4 * rand;
		end
		pAdjPlot = Map.GetAdjacentPlot(iX, iY, (direction + rotation) % 6);
		adjIndex = pAdjPlot:GetIndex();
		while protectionData[adjIndex].mustHaveOcean == false do
			protectionData[adjIndex].mustHaveOcean = true;
			pAdjPlot = Map.GetAdjacentPlot(iX, iY, (direction + rotation) % 6);
			adjIndex = pAdjPlot:GetIndex();
		end
		return;
	end
	
end

------------------------------------------------------------------------------

function AssimilateTerrain(index, validTerrains)
	
	local pPlot = Map.GetPlotByIndex(index);
	local terrainType = pPlot:GetTerrainType();
	
	-- Sea tiles
	if terrainType == g_TERRAIN_TYPE_COAST or terrainType == g_TERRAIN_TYPE_OCEAN then
		return g_TERRAIN_TYPE_COAST;
	end

	-- Get plot type
	local plotType = g_PLOT_TYPE_LAND;
	local terrainTypeFlat = terrainType;
	if pPlot:IsHills() then
		plotType = g_PLOT_TYPE_HILLS;
		terrainTypeFlat = terrainTypeFlat - 1;
	elseif pPlot:IsMountain() then
		plotType = g_PLOT_TYPE_MOUNTAIN;
		terrainTypeFlat = terrainTypeFlat - 2;
	end
	
	-- Get priority lists
	local aPlotTypes = g_aaPlotPriorities[plotType];
	local aTerrainTypes = g_aaTerrainPriorities[terrainTypeFlat];
	
	-- Find suitable terrain type
	local terrainTypeNew;
	for plotPriority, plotNew in pairs(aPlotTypes) do
		local terrainDelta = 0;
		if plotNew == g_PLOT_TYPE_HILLS then
			terrainDelta = 1;
		elseif plotNew == g_PLOT_TYPE_MOUNTAIN then
			terrainDelta = 2;
		end
		for terrainPriority, terrainNew in pairs(aTerrainTypes) do
			terrainTypeNew = terrainNew + terrainDelta;
			if validTerrains[terrainTypeNew] == true then
				break;
			end
		end
		if validTerrains[terrainTypeNew] == true then
			break;
		end
	end
	
	print("Converted terrain type ", terrainType, " to ", terrainTypeNew);
	
	return terrainTypeNew;
	
end

------------------------------------------------------------------------------

function ConfigureNaturalWonders(args)
	
	-- Unpack arguments
	local iW = args.iW;
	local iH = args.iH;
	local plotTypes = args.plotTypes;
	local protectionData = args.protectionData;
	local nwData = args.nwData;
	
	-- Terrain validity data
	local nwValidTerrains = {};
	local nwAdjacentTerrains = {};
	
	-- Count natural wonders
	local aNaturalWondersAvailable = {};
	local nNaturalWondersAvailable = 0;
	for feature in GameInfo.Features() do
		if feature.NaturalWonder then
			local featureType = g_FEATURE_BARRIER_REEF + nNaturalWondersAvailable;
			table.insert(aNaturalWondersAvailable, featureType);
			nwValidTerrains[featureType] = {};
			nwAdjacentTerrains[featureType] = {};
			for row in GameInfo.Terrains() do
				local terrainType = GetGameInfoIndex("Terrains", row.TerrainType);
				-- Assume no terrains are valid for placement
				nwValidTerrains[featureType][terrainType] = false;
				-- Assume all terrains are valid for adjacency
				nwAdjacentTerrains[featureType][terrainType] = true;
			end
			nNaturalWondersAvailable = nNaturalWondersAvailable + 1;
		end
	end
	
	-- Get valid terrain data
	for row in GameInfo.Feature_ValidTerrains() do
		local featureType = GetGameInfoIndex("Features", row.FeatureType);
		if IsNaturalWonderSupported(featureType) then
			local terrainType = GetGameInfoIndex("Terrains", row.TerrainType);
			nwValidTerrains[featureType][terrainType] = true;
		end
	end
	
	-- Get valid adjacent terrain data
	local abInvertAdjacentValidity = {};
	for row in GameInfo.Feature_AdjacentTerrains() do
		local featureType = GetGameInfoIndex("Features", row.FeatureType);
		if IsNaturalWonderSupported(featureType) then
			local terrainType = GetGameInfoIndex("Terrains", row.TerrainType);
			nwAdjacentTerrains[featureType][terrainType] = false;
			abInvertAdjacentValidity[featureType] = true;
		end
	end
	
	-- Invert NWs defined according to valid adjacent terrains (as opposed to invalid adjacent terrains)
	for featureType, bInvert in pairs(abInvertAdjacentValidity) do
		if bInvert == true then
			for terrainType, bAdjacent in pairs(nwAdjacentTerrains[featureType]) do
				nwAdjacentTerrains[featureType][terrainType] = not bAdjacent;
			end
		end
	end
	
	-- Get invalid adjacent terrain data
	for row in GameInfo.Feature_NotAdjacentTerrains() do
		local featureType = GetGameInfoIndex("Features", row.FeatureType);
		if IsNaturalWonderSupported(featureType) then
			local terrainType = GetGameInfoIndex("Terrains", row.TerrainType);
			nwAdjacentTerrains[featureType][terrainType] = false;
		end
	end
	
	-- Get natural wonders
	for i, data in ipairs(nwData) do
		local aCoastalTemp = {};
		while data.featureType == nil do
			local rand = 1 + TerrainBuilder.GetRandomNumber(nNaturalWondersAvailable, "Random Natural Wonder");
			local nw = aNaturalWondersAvailable[rand];
			table.remove(aNaturalWondersAvailable, rand);
			nNaturalWondersAvailable = nNaturalWondersAvailable - 1;
			if IsNaturalWonderSupported(nw) then
				if data.coastDir == -1 and IsNaturalWonderCoastal(nw) == true then
					table.insert(aCoastalTemp, nw);
				else
					data.featureType = nw;
					data.validTerrains = nwValidTerrains[nw];
					data.adjacentTerrains = nwAdjacentTerrains[nw];
				end
			end
		end
		while aCoastalTemp[1] ~= nil do
			table.insert(aNaturalWondersAvailable, aCoastalTemp[1]);
			nNaturalWondersAvailable = nNaturalWondersAvailable + 1;
			table.remove(aCoastalTemp, 1);
		end
		ApplyNaturalWonderProtection(plotTypes, protectionData, data);
	end
	
	return protectionData, nwData;
	
end

------------------------------------------------------------------------------

function IsNaturalWonderCoastal(nw)
	
	local aCoastalNaturalWonders = {
		g_FEATURE_BARRIER_REEF,
		g_FEATURE_CLIFFS_DOVER,
		g_FEATURE_GALAPAGOS,
		g_FEATURE_PIOPIOTAHI,
		g_FEATURE_LYSEFJORDEN,
		g_FEATURE_GIANTS_CAUSEWAY
	};
	
	for i, nwCoastal in ipairs(aCoastalNaturalWonders) do
		if nw == nwCoastal then
			return true
		end
	end
	
	return false;
	
end

------------------------------------------------------------------------------

function IsNaturalWonderSupported(nw)
	
	if nw == nil then
		return false;
	end
	
	if nw >= g_FEATURE_BARRIER_REEF and nw <= g_FEATURE_GIANTS_CAUSEWAY then
		return true;
	end
	
	return false;
	
end

------------------------------------------------------------------------------

-- "NV_MapUtil_Plots.lua" (to be separated)
-- VFS system currently does not allow inclusion of imported scripts

------------------------------------------------------------------------------

function AddHillsMountains(args)
	
	-- Unpack arguments
	local continentTypes = args.continentTypes;
	local plotTypes = args.plotTypes;
	local protectionData = args.protectionData;
	local heightRankings = args.heightRankings;
	local world_age = args.world_age;
	
	-- Get proportion of plots to promote to mountains
	local list_prop_mountains = {0.025, 0.050, 0.075};
	local prop_mountains = list_prop_mountains[args.world_age];
	if prop_mountains == nil then
		local rand = 1 + TerrainBuilder.GetRandomNumber(3, "Random Mountain Proportion");
		prop_mountains = list_prop_mountains[rand];
	end
	
	-- Get proportion of plots to promote to hills
	local list_prop_hills = {0.20, 0.40, 0.60};
	local prop_hills = list_prop_hills[args.world_age];
	if prop_hills == nil then
		local rand = 1 + TerrainBuilder.GetRandomNumber(3, "Random Hills Proportion");
		prop_hills = list_prop_hills[rand];
	end
	
	-- Add mountains and hills
	for continent, aRankings in ipairs(heightRankings) do
		local nMountains = math.floor(prop_mountains * aRankings.count);
		local nHills = math.floor(prop_hills * aRankings.count);
		local countMountains = 0;
		local countHills = 0;
		local listIndex = 1;
		local bDone = false;
		while bDone == false and aRankings[listIndex] ~= nil do
			local plotIndex = aRankings[listIndex];
			local pPlot = Map.GetPlotByIndex(plotIndex);
			local iX = pPlot:GetX();
			local iY = pPlot:GetY();
			if (
				countMountains < nMountains and
				protectionData[plotIndex].canHaveMountain == true and
				AdjacentToSaltWater(iX, iY) == false and
				IsIsthmus(plotTypes, iX, iY, g_PLOT_TYPE_MOUNTAIN) == false
			) then
				plotTypes[plotIndex] = g_PLOT_TYPE_MOUNTAIN;
				countMountains = countMountains + 1;
			elseif countHills < nHills then
				plotTypes[plotIndex] = g_PLOT_TYPE_HILLS;
				countHills = countHills + 1;
			end
			listIndex = listIndex + 1;
			if countMountains >= nMountains and countHills >= nHills then
				bDone = true;
			end
		end
	end
	
	return plotTypes;
	
end

------------------------------------------------------------------------------

function AddLakes(args)

	-- Unpack arguments
	local continentTypes = args.continentTypes;
	local plotTypes = args.plotTypes;
	local protectionData = args.protectionData;
	local heightRankings = args.heightRankings;
	local sea_level = args.sea_level;
	
	-- Get proportion of plots to transform to lakes
	local list_prop_lakes = {0.015, 0.030, 0.045};
	local prop_lakes = list_prop_lakes[sea_level];
	if prop_lakes == nil then
		local rand = 1 + TerrainBuilder.GetRandomNumber(3, "Random Lake Proportion");
		prop_lakes = list_prop_lakes[rand];
	end
	
	-- Add lakes
	local aLakePlots = {};
	for continent, aRankings in ipairs(heightRankings) do
		local nLakes = math.floor(prop_lakes * aRankings.count);
		local countLakes = 0;
		local listIndex = aRankings.count;
		-- Out-of-bounds check
		while listIndex > 0 and aRankings[listIndex] == nil do
			listIndex = listIndex - 1;
		end
		while countLakes < nLakes and aRankings[listIndex] ~= nil do
			local plotIndex = aRankings[listIndex];
			local pPlot = Map.GetPlotByIndex(plotIndex);
			local iX = pPlot:GetX();
			local iY = pPlot:GetY();
			if (
				protectionData[plotIndex].canHaveLake == true and
				pPlot:IsRiver() == false and
				pPlot:IsRiverAdjacent() == false and
				pPlot:IsCoastalLand() == false and
				IsIsthmus(plotTypes, iX, iY, g_PLOT_TYPE_OCEAN) == false and
				pPlot:IsNaturalWonder() == false
			) then
				plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
				table.insert(aLakePlots, plotIndex);
				countLakes = countLakes + 1;
			end
			listIndex = listIndex - 1;
		end
	end
	
	for k, index in ipairs(aLakePlots) do
		local pPlot = Map.GetPlotByIndex(index);
		TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_COAST);
	end
	
end

------------------------------------------------------------------------------

function ErodeLandmass(args)
	
	-- Unpack arguments
	local iW = args.iW;
	local iH = args.iH;
	local continentTypes = args.continentTypes;
	local plotTypes = args.plotTypes;
	local protectionData = args.protectionData;
	local aContinentModifiers = args.aContinentModifiers;
	local water_plot = args.water_plot;
	local water_terrain = args.water_terrain;
	local sea_level = args.sea_level;
	
	-- Get number of plots to erode
	local list_prop_ocean = {0.2, 0.4, 0.6};
	local prop_ocean = list_prop_ocean[args.sea_level];
	if prop_ocean == nil then
		local rand = 1 + TerrainBuilder.GetRandomNumber(3, "Random Ocean Proportion");
		prop_ocean = list_prop_ocean[rand];
	end
	
	-- Instantiate lists of plots
	local aaValidPlots = {};
	local anValidPlots = {};
	local anTotalPlots = {};
	for continent, modifier in ipairs(aContinentModifiers) do
		aaValidPlots[continent] = {};
		anValidPlots[continent] = 0;
		anTotalPlots[continent] = 0;
	end
	
	-- Populate plot lists
	for i = 0, (iW * iH) - 1 do
		local pPlot = Map.GetPlotByIndex(i);
		local continent = continentTypes[i];
		if continent > 0 and pPlot:IsWater() == false then
			anTotalPlots[continent] = anTotalPlots[continent] + 1;
			if protectionData[i].mustHaveOcean == true then
				-- Remove mandatory ocean plots (for natural wonder placement)
				plotTypes[i] = water_plot;
				TerrainBuilder.SetTerrainType(pPlot, water_terrain);
			elseif protectionData[i].canHaveOcean == true then
				-- Insert into list for every adjacent water plot
				local iX = pPlot:GetX();
				local iY = pPlot:GetY();
				for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
					local pAdjPlot = Map.GetAdjacentPlot(iX, iY, direction);
					if pAdjPlot ~= nil and pAdjPlot:IsWater() then
						table.insert(aaValidPlots[continent], i);
						anValidPlots[continent] = anValidPlots[continent] + 1;
					end
				end
			end
		end
	end
	
	-- Erode each continent
	for continent, validPlots in ipairs(aaValidPlots) do
		local nErosions = math.floor(prop_ocean * anTotalPlots[continent] * aContinentModifiers[continent]);
		for erosion = 1, nErosions do
			local pPlot;
			local iX;
			local iY;
			local index;
			local bFoundValidPlot = false;
			-- Find valid plot to erode
			while anValidPlots[continent] > 0 and not bFoundValidPlot do
				local rand = 1 + TerrainBuilder.GetRandomNumber(anValidPlots[continent], "Random Plot to Erode");
				index = validPlots[rand];
				pPlot = Map.GetPlotByIndex(index);
				iX = pPlot:GetX();
				iY = pPlot:GetY();
				table.remove(validPlots, rand);
				anValidPlots[continent] = anValidPlots[continent] - 1;
				if (plotTypes[index] ~= water_plot) and (IsIsthmus(plotTypes, iX, iY, water_plot) == false) then
					bFoundValidPlot = true;
				end
			end
			if bFoundValidPlot then
				plotTypes[index] = water_plot;
				TerrainBuilder.SetTerrainType(pPlot, water_terrain);
				-- Add adjacent plots to respective list
				for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
					local pAdjPlot = Map.GetAdjacentPlot(iX, iY, direction);
					local adjIndex = pAdjPlot:GetIndex();
					local adjContinent = continentTypes[adjIndex];
					if adjContinent ~= 0 and plotTypes[adjIndex] ~= water_plot and protectionData[adjIndex].canHaveOcean == true then
						table.insert(aaValidPlots[adjContinent], adjIndex);
						anValidPlots[adjContinent] = anValidPlots[adjContinent] + 1;
					end
				end
			end
		end
	end
	
	return plotTypes;
	
end

------------------------------------------------------------------------------

function GetHeightRankings(args)
	
	-- Unpack arguments
	local iW = args.iW;
	local iH = args.iH;
	local iFlags = args.iFlags;
	local continentTypes = args.continentTypes;
	local plotTypes = args.plotTypes;
	local grain_amount = args.grain_amount or 3;
	local fracXExp = args.fracXExp or -1;
	local fracYExp = args.fracYExp or -1;
	local numPlates = args.numPlates or 9;
	local blendRidge = args.blendRidge or 5;
	local blendFract = args.blendFract or 5;
	
	-- Generate fractal
	local frac = Fractal.Create(iW, iH, grain_amount, iFlags, fracXExp, fracYExp);
	frac:BuildRidges(numPlates, iFlags, blendRidge, blendFract);
	
	-- Populate lists
	local aaRankings = {};
	local aHeights = {};
	for i = 0, (iW * iH) - 1 do
		local pPlot = Map.GetPlotByIndex(i);
		if pPlot:IsWater() == false then
			local iX = pPlot:GetX();
			local iY = pPlot:GetY();
			aHeights[i] = frac:GetHeight(iX, iY);
			local continent = continentTypes[i];
			if aaRankings[continent] == nil then
				aaRankings[continent] = {};
				aaRankings[continent].count = 0;
			end
			table.insert(aaRankings[continent], i);
			aaRankings[continent].count = aaRankings[continent].count + 1;
		end
	end
	
	-- Sort continent plot lists
	for continent, aRankings in ipairs(aaRankings) do
		QuickSort(aRankings, aHeights, 1, aRankings.count);
	end
	
	return aaRankings;
	
end

------------------------------------------------------------------------------

-- "NV_MapUtil_Resources.lua" (to be separated)
-- VFS system currently does not allow inclusion of imported scripts

------------------------------------------------------------------------------

--[[
function AddResources(args)
	
	local luxuryDistribution = args.luxuryDistribution;
	
	-- Get list of luxuries
	local aLuxuries = {};
	for row in GameInfo.Resources() do
		if row.ResourceClassType == "RESOURCECLASS_LUXURY" then
			table.insert(aLuxuries, row.Index);
		end
	end
	
	-- Assign luxuries to each continent
	ShuffleArray(aLuxuries);
	for continent, distribution in pairs(luxuryDistribution) do
		for i, luxury in ipairs(distribution) do
			distribution[i] = aLuxuries[luxury];
		end
	end

end
]]

-------------------------------------------------------------------------------

-- "NV_MapUtil_Terrain.lua" (to be separated)
-- VFS system currently does not allow inclusion of imported scripts

-------------------------------------------------------------------------------

function ApplyTerrain(args)
	
	-- Unpack arguments
	local iW = args.iW;
	local iH = args.iH;
	local plotTypes = args.plotTypes;
	local terrainTypes = args.terrainTypes;
	
	print("Applying terrain");
	for i = 0, (iW * iH) - 1 do
		local pPlot = Map.GetPlotByIndex(i);
		if plotTypes[i] == g_PLOT_TYPE_HILLS then
			terrainTypes[i] = terrainTypes[i] + 1;
		end
		TerrainBuilder.SetTerrainType(pPlot, terrainTypes[i]);
	end
	print("Successfully applied terrain");
	
end
	
-------------------------------------------------------------------------------

function GenerateTerrain(args)
	
	-- Unpack arguments
	local iW = args.iW;
	local iH = args.iH;
	local iFlags = args.iFlags;
	local plotTypes = args.plotTypes;
	local temperature = args.temperature;
	
	-- Get temperature
	local list_temperature = {1, 2, 3};
	local temperature = list_temperature[temperature];
	if temperature == nil then
		local rand = 1 + TerrainBuilder.GetRandomNumber(3, "Random Temperature");
		temperature = list_temperature[rand];
	end
	
	-- Generate terrain
	local terrainTypes = GenerateTerrainTypes(plotTypes, iW, iH, iFlags, false, temperature, 0);
	
	return terrainTypes;
	
end

-------------------------------------------------------------------------------

-- "NV_MapUtil_Utilities.lua" (to be separated)
-- VFS system currently does not allow inclusion of imported scripts

-------------------------------------------------------------------------------

function IsIsthmus(propertyTypes, iX, iY, water_property)
	
	local pPlot = Map.GetPlot(iX, iY);
	local index = pPlot:GetIndex();
	
	-- Not an isthmus if water
	if propertyTypes[index] == water_property then
		return false;
	end
	
	-- Isthmus exists if cycling through adjacent plots reveals more than two changes between land and water
	local nWaterToggles = 0;
	local pAdjPlot = Map.GetAdjacentPlot(iX, iY, 5);
	local bAdjWater = propertyTypes[pAdjPlot:GetIndex()] == water_property;
	local pAdjNextPlot;
	local bAdjNextWater;
	for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
		pAdjNextPlot = Map.GetAdjacentPlot(iX, iY, direction);
		bAdjNextWater = propertyTypes[pAdjNextPlot:GetIndex()] == water_property;
		if bAdjWater ~= bAdjNextWater then
			nWaterToggles = nWaterToggles + 1;
		end
		bAdjWater = bAdjNextWater;
	end
	
	if nWaterToggles > 2 then
		return true;
	end
	
	return false;
	
end

-------------------------------------------------------------------------------

function QuickSort(arr, vals, iMin, iMax)
	
	-- Upper out-of-bounds check
	while iMin < iMax and vals[arr[iMax]] == nil do
		iMax = iMax - 1;
	end
	
	-- Lower out-of-bounds check
	while iMin < iMax and vals[arr[iMin]] == nil do
		iMin = iMin + 1;
	end
	
	-- Quicksort algorithm
	if iMin < iMax then
		local pivot = arr[iMax];
		local i = iMin;
		local j = iMax;
		while i < j do
			if vals[pivot] > vals[arr[i]] then
				arr[j] = arr[i];
				arr[i] = arr[j - 1];
				j = j - 1;
			else
				i = i + 1;
			end
		end
		arr[j] = pivot;
		QuickSort(arr, vals, iMin, j - 1);
		QuickSort(arr, vals, j + 1, iMax);
	end
	
end

-------------------------------------------------------------------------------

function ShuffleArray(arr)
	
	-- Count elements
	local count = 0;
	for k, v in ipairs(arr) do
		count = count + 1;
	end
	
	for i = 1, count - 1 do
		local rand = i + TerrainBuilder.GetRandomNumber(count - i + 1, "Random Element");
		local temp = arr[i];
		arr[i] = arr[rand];
		arr[rand] = temp;
	end
	
end

-------------------------------------------------------------------------------

function StampContinents(args)
	
	AreaBuilder.Recalculate();
	TerrainBuilder.AnalyzeChokepoints();
	
	-- Unpack arguments
	local continentTypes = args.continentTypes;
	
	local nContinentNamesAvailable = 0;
	local aContinentNamesAvailable = {};
	for continentName in GameInfo.Continents() do
		table.insert(aContinentNamesAvailable, nContinentNamesAvailable);
		nContinentNamesAvailable = nContinentNamesAvailable + 1;
	end
	
	local aContinentNameIds = {};
	for index, continent in pairs(continentTypes) do
		local pPlot = Map.GetPlotByIndex(index);
		if pPlot:IsWater() == false or pPlot:IsLake() then
			if aContinentNameIds[continent] == nil then
				local rand = TerrainBuilder.GetRandomNumber(nContinentNamesAvailable, "Random Continent Name");
				aContinentNameIds[continent] = aContinentNamesAvailable[rand];
				table.remove(aContinentNamesAvailable, rand);
				nContinentNamesAvailable = nContinentNamesAvailable - 1;
			end
			TerrainBuilder.SetContinentType(pPlot, aContinentNameIds[continent]);
		end
	end
	
end